"""
Symmetrization
--------------

"""

import numpy as np
import qutip as qt
from itertools import permutations, product

from spheres.utils import *

def symmetrize(pieces):
    """
    Given a list of quantum states, constructs their symmetrized
    tensor product. If given a multipartite quantum state instead, 
    we sum over all permutations on the subsystems.

    Parameters
    ----------
        pieces : list or qt.Qobj
            List of quantum states or a multipartite state.

    Returns
    -------
        sym : qt.Qobj
            Symmetrized tensor product.
    """
    if type(pieces) == list:
        return sum([qt.tensor(*[pieces[i] for i in perm])\
                for perm in permutations(range(len(pieces)))]).unit()
    else:
        return sum([pieces.permute(perm)\
                for perm in permutations(range(len(pieces.dims[0])))]).unit()
                
def spin_sym(j):
    """
    Constructs an isometric linear map from spin-j states to
    permutation symmetric states of 2j spin-:math:`\\frac{1}{2}`'s.

    Parameters
    ----------
        j : int
            j value.

    Returns
    -------
        S : qt.Qobj
            Linear map from :math:`2j+1` dimensions to :math:`2^{2j}` dimensions.

    """
    if j == 0:
        return qt.Qobj(1)
    S = qt.Qobj(np.vstack([\
                    components(symmetrize(\
                           [qt.basis(2,0)]*int(2*j-i)+\
                           [qt.basis(2,1)]*i))\
                for i in range(int(2*j+1))]).T)
    S.dims =[[2]*int(2*j), [int(2*j+1)]]
    return S

def symmetrized_basis(n, d=2):
    """
    Constructs a symmetrized basis set for n systems in d dimensions.

    Parameters
    ----------
        n : int
            The number of systems to symmetrize.

        d : int or list
            Either an integer representing the dimensionality
            of the individual subsystems, in which case, we work 
            in the computational basis; or else a list of basis
            states for the individual systems.

    Returns
    -------
        sym_basis : dict
            ``sym_basis["labels"]`` is a list of labels for
            the symmetrized basis states. Each element of the list
            is a tuple whose length is the dimensionality of the
            individual subsystems, with an integer counting the number
            of subsystems in that basis state.

            ``sym_basis["basis"]`` is a dictionary mapping 
            labels to symmetrized basis states.

            ``sym_basis["map"]`` is a linear transformation
            from the permutation symmetric subspace to the
            full tensor product of the n systems.

            The dimensionality of the symmetric subspace
            corresponds to the number of ways of distributing
            :math:`n` elements in :math:`d` boxes, where :math:`n` is the number of
            systems and :math:`d` is the dimensionality of
            an individual subsytems. In other words, the dimensionality
            :math:`s` of the permutation symmetric subspace is :math:`\\binom{d+n-1}{n}`.

            So ``sym_basis["map"]`` is a map from
            :math:`\\mathbb{C}^{s} \\rightarrow \\mathbb{C}^{d^{n}}`.

    """
    if type(d) == int:
        d = [qt.basis(d, i) for i in range(d)]
    labels = list(filter(lambda b: sum(b) == n,\
                list(product(range(n+1), repeat=len(d)))[::-1]))
    sym_basis = dict([(label, symmetrize(flatten([[d[i]]*b \
                                    for i, b in enumerate(label)])))\
                                        for label in labels])
    sym_map = qt.Qobj(np.vstack([components(sym_basis[label])\
                                        for label in labels]).T)
    sym_map.dims =[[d[0].shape[0]]*n, [len(sym_basis)]]
    return {"labels": labels,\
            "basis": sym_basis,\
            "map": sym_map}



"""
Utility Functions
-----------------
"""

import numpy as np
import qutip as qt

factorial = np.math.factorial

def rand_c():
    """
    Generates random extended complex coordinate whose real and imaginary parts
    are normally distributed, and ten percent of the time, we return :math:`\\infty`.
    """
    return np.random.randn() + 1j*np.random.randn() if np.random.random() > 0.1 else np.inf

def rand_xyz(n=1):
    """
    Generates n random points on the unit sphere in cartesian coordinates.
    """
    if n == 1:
        return normalize(np.random.randn(3))
    return np.array([normalize(np.random.randn(3)) for i in range(n)])

def rand_sph(unit=False):
    """
    Generates random spherical coordinate. If `unit=True`, the point
    lies on the surface of the sphere.
    """
    return np.array([1 if unit else np.random.random(),\
                     2*np.pi*np.random.random(),\
                     np.pi*np.random.random()])

def components(q):
    """
    Extracts components of qt.Qobj, whether bra or ket.
    """
    return (q.full().T[0] if q.type == "ket" else q.full()[0])\
                if type(q) == qt.Qobj else q

def normalize(v):
    """
    Normalizes numpy vector. Simply passes the vector through
    if its norm is 0.
    """
    n = np.linalg.norm(v)
    return v/n if not np.isclose(n, 0) else v

def phase(v):
    """
    Extracts phase of complex vector (np.ndarray or qt.Qobj) by finding
    the first non-zero component and returning its phase.
    """
    v = components(v) if type(v) == qt.Qobj else v
    return np.exp(1j*np.angle(v[(v!=0).argmax(axis=0)]))

def normalize_phase(v):
    """
    Normalizes the phase of a complex vector (np.ndarray or qt.Qobj).
    """
    return v/phase(v)

def compare_unordered(A, B, decimals=5):
    """
    Compares two sets of vectors regardless of their ordering.
    """
    A, B = np.around(A, decimals=decimals), np.around(B, decimals=decimals)
    return np.array([a in B for a in A]).all()

def compare_nophase(a, b):
    """
    Compares two vectors disregarding their overall complex phase.
    """
    return np.allclose(normalize_phase(a), normalize_phase(b))

def compare_spinors(A, B, decimals=5):
    """
    Compares two lists of spinors, disregarding both their phases,
    as well as their ordering in the list.
    """
    A = np.array([components(normalize_phase(a)) for a in A])
    B = np.array([components(normalize_phase(b)) for b in B])
    return compare_unordered(A, B, decimals=decimals)

def so3_generators():
    """
    Returns :math:`SO(3)` generators :math:`L_{x}, L_{y}, L_{z}`.
    """
    Lx = np.array([[0,0,0],\
                   [0,0,-1],\
                   [0,1,0]])
    Ly = np.array([[0,0,1],\
                   [0,0,0],\
                   [-1,0,0]])
    Lz = np.array([[0,-1,0],\
                   [1,0,0],\
                   [0,0,0]])
    return Lx, Ly, Lz

def bitstring_basis(bitstring, dims=2):
    """
    Generates a basis vector corresponding to a given bitstring, 
    which may be a list of integers or a string of integers. The
    dimensionality of each tensor factor is given by dims, which may
    be either an integer (all the same dimensionality) or a list 
    (a dimension for each factor).
    """
    if type(bitstring) == str:
        bitstring = [int(s) for s in bitstring]
    if type(dims) != list:
        dims = [dims]*len(bitstring)
    return qt.tensor(*[qt.basis(dims[i], i)\
                for i in range(len(bitstring))])

def fix_stars(old_stars, new_stars):
    """
    Try to adjust the ordering of a list of stars to keep continuity.
    """
    if np.all([np.allclose(old_stars[0], old_star) for old_star in old_stars]):
        return new_stars
    ordering = [None]*len(old_stars)
    for i, old_star in enumerate(old_stars):
        dists = np.array([np.linalg.norm(new_star-old_star) for new_star in new_stars])
        minim = np.argmin(dists)
        if np.count_nonzero(dists == dists[minim]) == 1:
            ordering[i] = new_stars[minim]
        else:
            return new_stars
    return ordering

def flatten(to_flatten):
    """
    Flatten list of lists.
    """
    return [item for sublist in to_flatten for item in sublist]

def binomial(n, k):
    """
    Returns the binomial coefficient :math:`\\binom{n}{k} = \\frac{n!}{k!(n-k)!}`
    """
    return int(factorial(n)/(factorial(k)*factorial(n-k)))



"""
Visualization Tools
-------------------

"""

from spheres.stars import *
from spheres.utils import *

import vpython as vp
global_scene = None

def tangent_plane_rotation(phi, theta):
    """
    Constructs rotation into the tangent plane to the sphere 
    at the given point specified in spherical coordinates. Returns
    a 3x3 np.ndarray representing the linear transformation corresponding
    to the rotation.

    """
    normal = sph_xyz(np.array([1, phi, theta]))
    tangent = sph_xyz(np.array([1, phi, theta+np.pi/2]))
    return np.linalg.inv(\
                np.array([tangent,\
                          normalize(np.cross(tangent, normal)),\
                          normal]))

class MajoranaSphere:
    """
    `MajoranaSphere` provides a nice way to visualize (pure) spin-j states using
    vpython for graphics, whether in a jupyter notebook or in a standalone 
    environment.

    Create a `MajoranaSphere` by specifying a spin state. The radius of the sphere is
    determined by the j value of the spin. One can
    specify a vpython scene in which to place the MajoranaSphere (if none is provided,
    one is created by default) the position at which to place the sphere, 
    the color and opacity of the sphere, the colors of the stars, whether the stars 
    leave trails, whether the sphere and the stars
    are draggable by mouse, whether to show the expected rotation axis of the
    spin as a yellow arrow, whether to show reference cartesian axes in red, whether
    to show the complex phase of the spin as a green arrow hovering atop the sphere, 
    and whether to additionally visualize the spin as a wavefunction on the sphere:
    both as a coherent state wave function or in terms of the Majorana function. 
    One can specify the number of sample points to evaluate at, and the wavefunction
    amplitudes are visualized as arrows tangent to the sphere at these points.

    Attributes
    ----------
    spin : qt.Qobj
        The spin-j state represented. If this attribute is set, the visualization
        is automatically updated.
    j : float
        Its j value.
    xyz : np.ndarray
        Majorana points in cartesian coordinates.
    phase : complex
        Complex phase of the spin state.
    show_rotation_axis : bool
        Whether to show the expected rotation axis. If this attribute is set, the visualization
        is automatically updated. Can also use :py:meth:`toggle_rotation_axis`.
    show_phase : bool
        Whether to show the phase. If this attribute is set, the visualization
        is automatically updated. Can also use :py:meth:`toggle_phase`.
    show_axes : bool
        Whether to show reference cartesian axes. If this attribute is set, the visualization
        is automatically updated. Can also use :py:meth:`toggle_axes`.
    sphere_draggable : bool
        Whether one can drag the sphere with the mouse. If this attribute is set, the visualization
        is automatically updated. Can also use :py:meth:`toggle_sphere_draggable`.
    stars_draggable : bool
        Whether one can drag the individual stars with the mouse. If this attribute is set, the visualization
        is automatically updated. Can also use :py:meth:`toggle_stars_draggable`.
    phase_draggable : bool
        Whether one can drag the phase around with the mouse. If this attribute is set, the visualization
        is automatically updated. Can also use :py:meth:`toggle_phase_draggable`.
    make_trails : bool
        Whether the stars leave trails. If this attribute is set, the visualization
        is automatically updated. Can also use :py:meth:`toggle_trails`.
    show_wavefunction : bool or str
        Can specify "majorana", "coherent_state", or `None/False`. If this attribute is set, the visualization
        is automatically updated. Can also use :py:meth:`toggle_wavefunction`.

    """
    def __init__(self, spin,\
                       scene=None,\
                       position=vp.vector(0,0,0),\
                       sphere_color=vp.color.blue,\
                       sphere_opacity=0.3,\
                       star_colors="random",\
                       make_trails=True,\
                       sphere_draggable=True,\
                       stars_draggable=True,\
                       show_rotation_axis=True,\
                       show_phase=True,\
                       phase_draggable=True,\
                       show_axes=True,\
                       show_wavefunction=None,\
                       wavefunction_samples=15):
        global global_scene
        """
        Parameters
        ----------
        spin : qt.Qobj
            The spin state to visualize.
        scene : vp.canvas, optional
            `vpython` canvas in which to place the sphere. If none is provided,
            a new canvas is created.
        position : vp.vector, optional
            3D `vpython` vector coordinates at which to place the sphere.
        sphere_color : vp.vector, optional
            `vpython` vector representing the RGB color values for the sphere. 
        sphere_opacity : float, optional
            Specifies the opacity of the sphere. 
        star_colors : vp.vector or list or str, optional
            Specifies the color of the stars. If given a `vpython` vector representing
            RGB color values, all the stars are colored with that color. If given a list
            of such vectors, the stars are colored with those colors. If given "random",
            random colors are generated for each star. Note that if the stars are
            different colors, there is some performance overhead as a little extra work
            has to be done to keep track of which star is which: in other words, 
            :py:meth:`spin_xyz`  doesn't nail down the ordering of the stars and so they
            may "switch places." This is undetectable if the stars are all colored the same,
            but if they aren't, we use :py:meth:`fix_ordering` to try to keep continuity.
        make_trails : bool, optional
            If `True`, the stars leave (colored) trails behind them. Similarly to the above,
            we must use :py:meth:`fix_ordering` to keep continuity.
        sphere_draggable : bool, optional
            If `True`, one can click the sphere to (un)select it (it turns purple), and
            then drag it around the scene.
        stars_draggable : bool, optional
            If `True`, one can click the stars to un(select) them (they turn magenta),
            and then drag them around the surface of the sphere.
        show_rotation_axis : bool, optional
            If `True`, shows the expected spin axis in yellow.
        show_phase : bool, optional
            If `True`, shows the complex phase of the wavefunction as a green arrow
            within a black ring hovering above the sphere like a halo.
        phase_draggable : bool, optional
            If `True`, allows the complex phase to be dragged via the mouse.
        show_axes : bool, optional
            If `True`, shows reference cartesian axes in red.
        show_wavefunction : None or str, optional
            If "majorana", visualizes the normalized Majorana function on the sphere.
            If "coherent_state", visualizes the coherent state wavefunction on the sphere.
            The two are antipodal to each other. 
            The complex amplitudes appear as arrows tangent to the surface of the sphere
            at each sample point.
        wavefunction_samples : int, optional
            Number of sample points :math:`n` at which to evaluate the wavefunction. The total
            number of points is :math:`n^2`.

        """
        super().__setattr__("spin", spin)
        self.j = (self.spin.shape[0]-1)/2
        self.xyz = spin_xyz(self.spin)
        self.phase = phase(self.spin)
        
        if scene != None:
            self.scene = None
        else:
            if global_scene == None:
                global_scene = vp.canvas(background=vp.color.white,\
                                         align="center", 
                                         width=600, 
                                         height=600)
            self.scene = global_scene 
                        
        self.vsphere = vp.sphere(pos=position,\
                                 radius=self.j,\
                                 color=sphere_color,\
                                 opacity=sphere_opacity)

        self.fix_stars = True if make_trails == True else False
        if star_colors == None:
            self.star_colors = [vp.color.white]*len(self.xyz)
        elif type(star_colors) == vp.vector:
            self.star_colors = [star_colors]*len(self.xyz)
        elif type(star_colors) == list:
            self.star_colors = star_colors
            self.fix_stars = True
        elif star_colors == "random":
            self.star_colors = [vp.vector(*np.random.random(3)) for i in range(len(self.xyz))]
            self.fix_stars = True

        self.vstars = [vp.sphere(pos=self.vsphere.pos+self.vsphere.radius*vp.vector(*xyz),\
                                 color=self.star_colors[i],\
                                 radius=0.2*self.vsphere.radius,\
                                 make_trail=True,
                                 emissive=True) for i, xyz in enumerate(self.xyz)]

        self.toggle_rotation_axis(show_rotation_axis)
        self.toggle_phase(show_phase)
        self.toggle_axes(show_axes)
        self.toggle_sphere_draggable(sphere_draggable)
        self.toggle_stars_draggable(stars_draggable)
        self.toggle_phase_draggable(phase_draggable)
        self.toggle_trails(make_trails)

        self.wavefunction_samples = wavefunction_samples
        self.toggle_wavefunction(show_wavefunction)

        self.evolving = False
        self.saved_make_trails = None
        self.refreshing = False
        self.snapshots = []

    def __exists__(self, attr):
        return hasattr(self, attr) and getattr(self, attr) != None

    def __setattr__(self, name, value):
        if name == "show_rotation_axis":
            self.toggle_rotation_axis(value)
        elif name == "show_phase":
            self.toggle_phase(value)
        elif name == "show_axes":
            self.toggle_axes(value)
        elif name == "sphere_draggable":
            self.toggle_sphere_draggable(value)
        elif name == "stars_draggable":
            self.toggle_stars_draggable(value)
        elif name == "phase_draggable":
            self.toggle_phase_draggable(value)
        elif name == "make_trails":
            self.toggle_trails(value)
        elif name == "show_wavefunction":
            self.toggle_wavefunction(value)
        super().__setattr__(name, value)
        if name == "spin":
            self.clear_trails()
            self.refresh()

    def toggle_rotation_axis(self, toggle=None):
        super().__setattr__("show_rotation_axis", \
            (True if not self.show_rotation_axis else False) if toggle == None else toggle)
        if self.show_rotation_axis and not self.__exists__("vrotation_axis"):
            self.vrotation_axis = vp.arrow(pos=self.vsphere.pos,\
                                           color=vp.color.yellow,\
                                           axis=vp.vector(*sum([0.5*xyz for xyz in self.xyz])))
        if not self.show_rotation_axis and self.__exists__("vrotation_axis"):
            self.vrotation_axis.visible = False
            self.vrotation_axis = None    

    def toggle_phase(self, toggle=None):
        super().__setattr__("show_phase", \
            (True if not self.show_phase else False) if toggle == None else toggle)
        if self.show_phase and not self.__exists__("vphase"):
            self.vphase_ring = vp.ring(pos=self.vsphere.pos+vp.vector(0,self.j+0.1,0),\
                                       radius=self.j,\
                                       thickness=0.03*self.j,\
                                       axis=vp.vector(0,1,0),\
                                       color=vp.color.black,\
                                       opacity=0.4)
            self.vphase = vp.arrow(pos=self.vphase_ring.pos,\
                                   axis=self.j*vp.vector(self.phase.real,0,self.phase.imag),\
                                   shaftwidth=self.j*0.05,\
                                   opacity=0.3,\
                                   color=vp.color.green)
        if not self.show_phase and self.__exists__("vphase"):
            self.vphase_ring.visible = False
            self.vphase.visible = False
            self.vphase_ring = None
            self.vphase = None

    def toggle_axes(self, toggle=None):
        super().__setattr__("show_axes", \
            (True if not self.show_axes else False) if toggle == None else toggle)
        if self.show_axes and not self.__exists__("vaxes"):
            self.vaxes = [vp.arrow(pos=self.vsphere.pos,\
                                   axis=1.5*self.j*axis,\
                                   color=vp.color.red,\
                                   shaftwidth=0.05*self.j,\
                                   opacity=0.4) for axis in [vp.vector(1,0,0),\
                                                             vp.vector(0,1,0),\
                                                             vp.vector(0,0,1)]]
            self.vaxis_labels = [vp.text(text=label,\
                                         color=vp.color.red,\
                                         height=self.j*0.2,\
                                         pos=self.vsphere.pos+self.vaxes[i].axis)\
                                            for i, label in enumerate(["X", "Y", "Z"])]
        if not self.show_axes and self.__exists__("vaxes"):
            for axis in self.vaxes:
                axis.visible = False
            for label in self.vaxis_labels:
                label.visible = False
            self.vaxes = None
            self.vaxis_labels = None

    def toggle_sphere_draggable(self, toggle=None):
        super().__setattr__("sphere_draggable", \
            (True if not self.sphere_draggable else False) if toggle == None else toggle)
        if self.sphere_draggable:
            self.sphere_dragging = False
            if not self.__exists__("clicks_bound"):
                self.clicks_bound = True 
                self.scene.bind('click', self.mouseclick)
                self.scene.bind('mousemove', self.mousemove)
        if not self.sphere_draggable:
            self.sphere_dragging = None
            self.vsphere.color = vp.color.blue
            if self.__exists__("clicks_bound") and not self.stars_draggable and not self.phase_draggable:
                self.clicks_bound = None
                self.scene.unbind('click', self.mouseclick)
                self.scene.unbind('mousemove', self.mousemove)

    def toggle_stars_draggable(self, toggle=None):
        super().__setattr__("stars_draggable", \
            (True if not self.stars_draggable else False) if toggle == None else toggle)
        if self.stars_draggable:
            self.star_dragging = -1
            if not self.__exists__("clicks_bound"):
                self.clicks_bound = True 
                self.scene.bind('click', self.mouseclick)
                self.scene.bind('mousemove', self.mousemove)
        if not self.stars_draggable:
            self.star_dragging = -1
            for i, star in enumerate(self.vstars):
                star.color = self.star_colors[i]
            if self.__exists__("clicks_bound") and not self.sphere_draggable and not self.phase_draggable:
                self.clicks_bound = None
                self.scene.unbind('click', self.mouseclick)
                self.scene.unbind('mousemove', self.mousemove)

    def toggle_phase_draggable(self, toggle=None):
        super().__setattr__("phase_draggable", \
            (True if not self.phase_draggable else False) if toggle == None else toggle)
        if self.phase_draggable:
            self.phase_dragging = False
            if not self.__exists__("clicks_bound"):
                self.clicks_bound = True 
                self.scene.bind('click', self.mouseclick)
                self.scene.bind('mousemove', self.mousemove)
        if not self.phase_draggable:
            self.phase_dragging = None
            if self.__exists__("clicks_bound") and not self.sphere_draggable and not self.stars_draggable:
                self.clicks_bound = None 
                self.scene.unbind('click', self.mouseclick)
                self.scene.unbind('mousemove', self.mousemove)

    def toggle_trails(self, toggle=None):
        super().__setattr__("make_trails", \
            (True if not self.make_trails else False) if toggle == None else toggle)
        for star in self.vstars:
            star.clear_trail()
            star.make_trail = self.make_trails

    def clear_trails(self):
        """
        Clears the star trails.
        """
        for star in self.vstars:
            star.clear_trail()

    def toggle_wavefunction(self, toggle=None):
        if toggle != None and self.__exists__("show_wavefunction"):
            self.toggle_wavefunction()
        super().__setattr__("show_wavefunction", toggle)
        if self.show_wavefunction == None or self.show_wavefunction == False:
            if self.__exists__("vwavefunction"):
                for i in range(self.wavefunction_samples):
                    for j in range(self.wavefunction_samples):
                        self.vwavefunction[i][j].visible = False
                self.vwavefunction = None
        else:
            self.phi, self.theta = \
                np.meshgrid(
                    np.linspace(0, 2*np.pi, self.wavefunction_samples),\
                    np.linspace(0, np.pi, self.wavefunction_samples))
            self.tangent_plane_rotations = \
                    [[tangent_plane_rotation(self.phi[i][j], self.theta[i][j])\
                        for j in range(self.wavefunction_samples)]
                            for i in range(self.wavefunction_samples)]
            if self.show_wavefunction == "coherent_state":
                self.coherent_states = \
                    [[qt.spin_coherent(self.j, self.theta[i][j], self.phi[i][j])\
                            for j in range(self.wavefunction_samples)]\
                                for i in range(self.wavefunction_samples)]
                amps = [[(self.coherent_states[i][j].dag()*self.spin)[0][0][0]\
                                for j in range(self.wavefunction_samples)]\
                                    for i in range(self.wavefunction_samples)]
            elif self.show_wavefunction == "majorana":
                self.poly = spin_poly(self.spin, spherical=True, normalized=True)
                amps = [[self.poly(np.array([1, self.phi[i][j], self.theta[i][j]]))\
                                for j in range(self.wavefunction_samples)]\
                                    for i in range(self.wavefunction_samples)]
            self.vwavefunction = []
            for i in range(self.wavefunction_samples):
                row = []
                for j in range(self.wavefunction_samples):
                    arrow = vp.arrow(pos=self.vsphere.pos+self.vsphere.radius*vp.vector(*sph_xyz([1, self.phi[i][j], self.theta[i][j]])),\
                                     opacity=0.4)
                    axis = self.j*0.5*(self.tangent_plane_rotations[i][j] @ np.array([amps[i][j].real, amps[i][j].imag,0]))
                    if np.isclose(np.linalg.norm(axis), 0):
                        arrow.visible = False
                    arrow.axis = vp.vector(*axis)
                    row.append(arrow)
                self.vwavefunction.append(row)

    def mouseclick(self):
        if self.phase_draggable:
            if self.scene.mouse.pick == self.vphase:
                self.phase_dragging = True if not self.phase_dragging else False
            else:
                self.phase_dragging = False
            if self.phase_dragging:
                self.vphase.color = vp.color.magenta
            else:
                self.vphase.color = vp.color.green

        if self.sphere_draggable:
            if self.scene.mouse.pick == self.vsphere:
                self.sphere_dragging = True if not self.sphere_dragging else False
            else:
                self.sphere_dragging = False

            if self.sphere_dragging:
                self.vsphere.color = vp.color.purple
                self.saved_make_trails = self.make_trails
                self.toggle_trails(False)
            else:
                self.vsphere.color = vp.color.blue
                if self.saved_make_trails != None:
                    self.toggle_trails(self.saved_make_trails)
                    self.saved_make_trails = None

        if self.stars_draggable and not self.evolving:
            pick = self.scene.mouse.pick 
            if pick in self.vstars:
                index = self.vstars.index(pick)
                self.star_dragging = index if self.star_dragging != index else -1
            else:
                self.star_dragging = -1

            if self.star_dragging != -1:
                self.saved_make_trails = self.make_trails
                self.toggle_trails(False)
                for i, vstar in enumerate(self.vstars):
                    if i == self.star_dragging:
                        vstar.color = vp.color.magenta
                    else:
                        vstar.color = self.star_colors[i]
            else:
                for i, vstar in enumerate(self.vstars):
                    vstar.color = self.star_colors[i]
                if self.saved_make_trails != None:
                    self.toggle_trails(self.saved_make_trails)
                    self.saved_make_trails = None

    def mousemove(self):
        if self.show_phase:
            if self.phase_dragging:
                x, y, z = self.scene.mouse.pos.value
                x, z = normalize(np.array([x, z]))
                phase = x + 1j*z
                new_spin = phase*normalize_phase(self.spin)
                super().__setattr__("spin", new_spin)
                if not self.evolving and not self.refreshing:
                    self.refresh()
                return

        if self.stars_draggable and not self.evolving:
            if self.star_dragging != -1:
                xyz = normalize(np.array((self.scene.mouse.pos-self.vsphere.pos).value))
                self.xyz[self.star_dragging] = xyz
                new_spin = self.phase*xyz_spin(self.xyz)
                super().__setattr__("spin", new_spin)
                if not self.evolving and not self.refreshing:
                    self.refresh()
                return

        if self.sphere_draggable:
            if self.sphere_dragging:
                self.vsphere.pos = self.scene.mouse.pos
                if not self.evolving and not self.refreshing:
                    self.refresh()

    def snapshot(self):
        """
        Takes a snapshot of the current locations of the stars
        and, if applicable, the phase, rotation axis, and
        wavefunction amplitudes. In other words, clones them.
        Can take multiple snapshots, which are added to a list.
        """
        clone = {}
        clone["vstars"] = [star.clone() for star in self.vstars]
        if self.show_phase:
            clone["vphase"] = self.vphase.clone()
        if self.show_rotation_axis:
            clone["vrotation_axis"] = self.vrotation_axis.clone()
        if self.show_wavefunction:
            clone["vwavefunction"] = [[self.vwavefunction[i][j].clone()
                        for j in range(self.wavefunction_samples)]\
                            for i in range(self.wavefunction_samples)]
        self.snapshots.append(clone)

    def clear_snapshot(self):
        """
        Clears a snapshot. If there's more than one snapshot,
        we clear them first in, first out.
        """
        clone = self.snapshots.pop(0)
        for star in clone["vstars"]:
            star.visible = False
        if "vphase" in clone:
            clone["vphase"].visible = False
        if "vrotation_axis" in clone:
            clone["vrotation_axis"].visible = False
        if "vwavefunction" in clone:
            for i in range(self.wavefunction_samples):
                for j in range(self.wavefunction_samples):
                    clone["vwavefunction"][i][j].visible = False
        clone = {}

    def refresh(self):
        """
        Refreshes the visualization from the value of `self.spin`.
        """
        self.refreshing = True
        self.xyz = spin_xyz(self.spin) if not self.fix_stars else fix_stars(self.xyz, spin_xyz(self.spin))
        self.phase = phase(self.spin)

        for i, xyz in enumerate(self.xyz):
            self.vstars[i].pos = self.vsphere.pos + self.vsphere.radius*vp.vector(*xyz)

        if self.show_rotation_axis:
            self.vrotation_axis.pos = self.vsphere.pos
            self.vrotation_axis.axis = vp.vector(*sum([0.5*xyz for xyz in self.xyz]))

        if self.show_phase:
            self.vphase_ring.pos = self.vsphere.pos + vp.vector(0, self.j+0.1, 0)
            self.vphase.pos = self.vphase_ring.pos
            self.vphase.axis = self.j*vp.vector(self.phase.real, 0, self.phase.imag)

        if self.show_axes:
            for i, axis in enumerate(self.vaxes):
                axis.pos = self.vsphere.pos
                self.vaxis_labels[i].pos = self.vsphere.pos + axis.axis

        if self.show_wavefunction:
            if self.show_wavefunction == "coherent_state":
                amps = [[(self.coherent_states[i][j].dag()*self.spin)[0][0][0]\
                                for j in range(self.wavefunction_samples)]\
                                    for i in range(self.wavefunction_samples)]
            elif self.show_wavefunction == "majorana":
                self.poly = spin_poly(self.spin, spherical=True, normalized=True)
                amps = [[self.poly(np.array([1, self.phi[i][j], self.theta[i][j]]))\
                                for j in range(self.wavefunction_samples)]\
                                    for i in range(self.wavefunction_samples)]                   

            for i in range(self.wavefunction_samples):
                for j in range(self.wavefunction_samples):
                    axis = self.j*0.5*(self.tangent_plane_rotations[i][j] @ np.array([amps[i][j].real, amps[i][j].imag, 0]))
                    self.vwavefunction[i][j].visible = False if np.isclose(np.linalg.norm(axis),0) else True
                    self.vwavefunction[i][j].axis = vp.vector(*axis)
                    if self.sphere_dragging:
                        self.vwavefunction[i][j].pos = self.vsphere.pos+self.vsphere.radius*vp.vector(*sph_xyz([1, self.phi[i][j], self.theta[i][j]]))
        self.refreshing = False

    def evolve(self, H, dt=0.05, T=2*np.pi):
        """
        Visualizes the evolution of the spin state under the specified Hamiltonian.

        Parameters
        ----------
        H : qt.Qobj
            The Hamiltonian. 
        dt : float
            The time step for each sample of the evolution after which the 
            visualization is updated.
        T : int
            The total number of time steps to evolve for.
        """
        self.evolving = True
        U = (-1j*H*dt).expm()
        for t in np.linspace(0, T, int(T/dt)):
            super().__setattr__("spin", U*self.spin)
            self.refresh()
            vp.rate(50)
        self.evolving = False

    def destroy(self):
        """
        Destroys the MajoranaSphere.
        """
        self.vsphere.visible = False
        self.vsphere = None
        for star in self.vstars:
            star.clear_trail()
            star.visible = False
        self.vstars = []
        if self.show_rotation_axis:
            self.vrotation_axis.visible = False
            self.vrotation_axis = None
        if self.show_phase:
            self.vphase_ring.visible = False
            self.vphase_ring = None
            self.vphase.visible = False
            self.vphase = None
        if self.show_axes:
            for i, axis in enumerate(self.vaxes):
                axis.visible = False
                self.vaxis_labels[i].visible = False
            self.vaxes = None
            self.vaxis_labels = None
        if self.show_wavefunction:
            for i in range(self.wavefunction_samples):
                for j in range(self.wavefunction_samples):
                    self.vwavefunction[i][j].visible = False
            self.vwavefunction = None
        while len(self.snapshots) > 0:
            self.clear_snapshot()




"""
..  automodule:: spheres.stars
	:members:

..  automodule:: spheres.symmetrization
	:members:

.. automodule:: spheres.visualization
	:members:

.. automodule:: spheres.utils
	:members:
	
"""

import numpy as np
import qutip as qt

from spheres import utils
from spheres.utils import *

from spheres import stars
from spheres.stars import *

from spheres import symmetrization
from spheres.symmetrization import *

from spheres import visualization
from spheres.visualization import *

__version__ = "0.3.0.5"
 


import setuptools

with open("README.rst", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="spheres",
    version="0.3.0.5",
    author="Matthew Weiss",
    author_email="heyredhat@gmail.com",
    description="toolbox for higher spin and symmetrization",
    long_description=long_description,
    long_description_content_type="text/x-rst",
    url="https://github.com/heyredhat/spheres",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    install_requires = ['numpy', 'qutip', 'matplotlib', 'vpython']
)


"""
Majorana Stars
--------------

Implementation of the "Majorana stars" formalism for higher spin.

+-----------------------+-------------------------------+
| :py:meth:`c_xyz`      | Extended complex to cartesian |
| :py:meth:`xyz_c`      | and back.                     |
+-----------------------+-------------------------------+
| :py:meth:`xyz_sph`    | Cartesian to spherical        |
| :py:meth:`sph_xyz`    | and back.                     |
+-----------------------+-------------------------------+
| :py:meth:`c_sph`      | Extended complex to spherical |
| :py:meth:`sph_c`      | and back.                     |
+-----------------------+-------------------------------+
| :py:meth:`c_spinor`   | Extended complex to spinor    |
| :py:meth:`spinor_c`   | and back.                     |
+-----------------------+-------------------------------+
| :py:meth:`xyz_spinor` | Cartesian to spinor           |
| :py:meth:`spinor_xyz` | and back.                     |
+-----------------------+-------------------------------+
| :py:meth:`spinor_sph` | Spinor to spherical           |
| :py:meth:`sph_spinor` | and back.                     |
+-----------------------+-------------------------------+

+-------------------------+------------------------+
| :py:meth:`spin_poly`    | Spin to polynomial     |
| :py:meth:`poly_spin`    | and back.              |
+-------------------------+------------------------+
| :py:meth:`poly_roots`   | Polynomial to extended |
| :py:meth:`roots_poly`   | complex roots and back.|
+-------------------------+------------------------+
| :py:meth:`spin_xyz`     | Spin to cartesian roots|
| :py:meth:`xyz_spin`     | and back.              |
+-------------------------+------------------------+
| :py:meth:`spin_spinors` | Spin to spinorial roots|
| :py:meth:`spinors_spin` | and back.              |
+-------------------------+------------------------+
| :py:meth:`spin_c`       | Spin to extended       |
| :py:meth:`c_spin`       | complex roots and back.|
+-------------------------+------------------------+
| :py:meth:`spin_sph`     | Spin to spherical roots|
| :py:meth:`sph_spin`     | and back.              |
+-------------------------+------------------------+

+--------------------------+------------------------+
| :py:meth:`spin_coherent` | Spin coherent state.   |
+--------------------------+------------------------+
| :py:meth:`antipodal`     | Invert coordinates     |
|                          | on sphere.             |
+--------------------------+------------------------+
| :py:meth:`poleflip`      | Flip stereographic     |
|                          | projection pole.       |
+--------------------------+------------------------+
| :py:meth:`mobius`        | Construct Möbius       |
|                          | transformation.        |
+--------------------------+------------------------+
|:py:meth:`xyz_eigenstates`| Construct XYZ          |
|                          | eigenstates.           |
+--------------------------+------------------------+

"""

import numpy as np
import qutip as qt

from itertools import combinations

from spheres.utils import *

def c_xyz(c, pole="south"):
    """
    `Stereographic projection <https://en.wikipedia.org/wiki/Stereographic_projection>`_ 
    from the extended complex plane to the unit sphere. Given coordinate :math:`c=x+iy` or :math:`\\infty`:

        | If :math:`c = \\infty`, returns :math:`(0,0,-1)`.
        | Otherwise, returns :math:`(\\frac{2x}{1+x^2+y^2}, \\frac{2y}{1+x^2+y^2}, \\frac{1-x^2-y^2}{1+x^2+y^2})`.

    Parameters
    ----------
    c : complex or inf
        Point on the extended complex plane.
    pole : str, default 'south'
        Whether to project from the North or South pole.

    Returns
    -------
    np.ndarray
        Cartesian coordinates of point on unit sphere.

    """
    if(pole == "south"):
        if c == np.inf:
            return np.array([0,0,-1])
        else:
            x, y = c.real, c.imag
            return np.array([2*x/(1 + x**2 + y**2),\
                             2*y/(1 + x**2 + y**2),\
                   (1-x**2-y**2)/(1 + x**2 + y**2)])
    elif (pole == "north"):
        if c == np.inf:
            return np.array([0,0,1])
        else:
            x, y = c.real, c.imag
            return np.array([2*x/(1 + x**2 + y**2),\
                             2*y/(1 + x**2 + y**2),\
                   (-1+x**2+y**2)/(1 + x**2 + y**2)])

def xyz_c(xyz, pole="south"):
    """
    Reverse `Stereographic projection <https://en.wikipedia.org/wiki/Stereographic_projection>`_ 
    from the unit sphere to the extended complex plane.

        | Given :math:`(0,0,-1)`, returns :math:`\\infty`.
        | Otherwise returns :math:`c = (\\frac{x}{1+z}) + i(\\frac{y}{1+z})`.

    Parameters
    ----------
    xyz : np.ndarray
        Cartesian coordinates of point on unit sphere.
    pole : str, default 'south'
        Whether to reverse project from the North or South pole.

    Returns
    -------
    complex or inf
        Extended complex coordinate.

    """
    x, y, z = xyz
    if pole == "south":
        if np.isclose(z, -1):
            return np.inf
        else:
            return x/(1+z) + 1j*y/(1+z)
    elif pole == "north":
        if np.isclose(z, 1):
            return np.inf
        else:
            return x/(1-z) + 1j*y/(1-z)

def xyz_sph(xyz):
    """
    Converts cartesian coordinates :math:`(x, y, z)` to spherical coordinates :math:`(r, \\phi, \\theta)`.
    We use the physicist's convention: 

        | radius :math:`r = \\sqrt{x^2 + y^2 + z^2} \\in [0, \\infty)`
        | azimuth :math:`\\phi = \\arctan{\\frac{y}{x}} \\in [0, 2\\pi]`
        | inclination: :math:`\\theta = \\arccos{\\frac{z}{\\sqrt{x^2 + y^2 + z^2}}} \\in [0, \\pi]`

    Parameters
    ----------
    xyz : np.ndarray
        Cartesian coordinates.

    Returns
    -------
    sph : np.ndarray
        Spherical coordinates.
    """
    x, y, z = xyz
    return np.array([np.sqrt(x**2 + y**2 + z**2),\
                     np.mod(np.arctan2(y, x), 2*np.pi),\
                     np.arccos(z/np.sqrt(x**2 + y**2 + z**2))])

def sph_xyz(sph):
    """
    Converts spherical coordinates :math:`(r, \\phi, \\theta)` to cartesian coordinates :math:`(x, y, z)`.
    We use the physicist's convention: :math:`r \\in [0, \\infty)`, :math:`\\phi \\in [0, 2\\pi]`, :math:`\\theta \\in [0, \\pi]`.
    
        | :math:`x = r\\sin{\\theta}\\cos(\\phi)`
        | :math:`y = r\\sin{\\theta}\\sin(\\phi)`
        | :math:`z = r\\cos(\\theta)`

    Parameters
    ----------
    sph : np.ndarray
        Spherical coordinates.

    Returns
    -------
    xyz : np.ndarray
        Cartesian coordinates.
    """
    r, phi, theta = sph
    return np.array([r*np.sin(theta)*np.cos(phi),\
                     r*np.sin(theta)*np.sin(phi),\
                     r*np.cos(theta)])

def c_sph(c):
    """
    Converts extended complex coordinate to spherical coordinates.

    Parameters
    ----------
    c : complex or inf
        Extended complex coordinate.

    Returns
    -------
    sph : np.ndarray
        Spherical coordinates :math:`r, \\phi, \\theta`.
    """
    return xyz_sph(c_xyz(c))

def sph_c(sph):
    """
    Converts spherical coordinates to extended complex coordinate.

    Parameters
    ----------
    sph : np.ndarray
        Spherical coordinates :math:`r, \\phi, \\theta`.

    Returns
    -------
    c : complex or inf
        Extended complex coordinate.
    """
    return xyz_c(sph_xyz(sph))

def c_spinor(c):
    """
    Converts extended complex coordinate to a spinor.

        | If :math:`c = \\infty`, returns :math:`\\begin{pmatrix} 0 \\\\ 1 \\end{pmatrix}`.
        | Otherwise, returns :math:`\\frac{1}{\\sqrt{1+|c|^2}} \\begin{pmatrix} 1 \\\\ c \\end{pmatrix}`
    
    Parameters
    ----------
    c : complex or inf
        Extended complex coordinate.

    Returns
    -------
    spinor : qt.Qobj
        Normalized spinor.
    """
    return qt.Qobj(np.array([0, 1])) if c == np.inf \
        else (1/np.sqrt(1+np.abs(c)**2))*qt.Qobj(np.array([1, c]))

def spinor_c(spinor):
    """
    Converts spinor :math:`\\begin{pmatrix} a \\\\ b \\end{pmatrix}` to extended complex coordinate.

        | If :math:`a = 0`, returns :math:`\\infty`. 
        | Otherwise returns :math:`\\frac{b}{a}`.

    Parameters
    ----------
    spinor : qt.Qobj
        Normalized spinor.

    Returns
    -------
    c : complex or inf
        Extended complex coordinate.
    """
    a, b = components(spinor)
    return b/a if not np.isclose(a, 0) else np.inf

def xyz_spinor(xyz):
    """
    Converts cartesian coordinates to a spinor by reverse stereographic projection
    to the extended complex plane, and then lifting the latter to a 2-vector. 

    Parameters
    ----------
    xyz : np.ndarray
        Cartesian coordinates.

    Returns
    -------
    spinor : qt.Qobj
        Spinor.
    """
    return c_spinor(xyz_c(xyz))

def spinor_xyz(spinor):
    """
    Converts spinor :math:`\\mid \\psi \\rangle` to cartesian coordinates by taking the expectation values
    with the three Pauli matrices: :math:`(\\langle \\psi \\mid X \\mid \\psi \\rangle, \\langle \\psi \\mid Y \\mid \\psi \\rangle, \\langle \\psi \\mid Z \\mid \\psi \\rangle)`.

    Parameters
    ----------
    spinor : qt.Qobj
        Spinor.

    Returns
    -------
    xyz : np.ndarray
        Cartesian coordinates.
    """
    return np.array([qt.expect(qt.sigmax(), spinor),\
                     qt.expect(qt.sigmay(), spinor),\
                     qt.expect(qt.sigmaz(), spinor)])

def spinor_sph(spinor):
    """
    Converts spinor to spherical coordinates. 

    Parameters
    ----------
    spinor : qt.Qobj
        Spinor.

    Returns
    -------
    sph : np.ndarray
        Spherical coordinates :math:`r, \\phi, \\theta`.
    """
    return xyz_sph(spinor_xyz(spinor))

def sph_spinor(sph):
    """
    Converts spherical coordinates to spinor. 

    Parameters
    ----------
    sph : np.ndarray
        Spherical coordinates :math:`r, \\phi, \\theta`.

    Returns
    -------
    spinor : qt.Qobj
        Spinor.
    """
    return xyz_spinor(sph_xyz(sph))

def spin_poly(spin, projective=False,\
                    homogeneous=False,\
                    cartesian=False,\
                    spherical=False,\
                    normalized=False):
    """
    Converts a spin into its Majorana polynomial, which is defined as follows:

    .. math::

        p(z) = \\sum_{m=-j}^{m=j} (-1)^{j+m} \\sqrt{\\frac{(2j)!}{(j-m)!(j+m)!}} a_{j+m} z^{j-m}

    Here, the :math:`a`'s run through the components of the spin in the :math:`\\mid j, m\\rangle` representation. 
    Note that :math:`\\frac{(2j)!}{(j-m)!(j+m)!}` amounts to: :math:`\\binom{2j}{j+m}`, a binomial coefficient.

    By default, returns the coefficients of the Majorana polynomial as an np.ndarray.

    If `projective=True`, returns a function which takes an extended complex coordinate
    as an argument, and which evaluates the polynomial at that point. Note that to evaluate the polynomial
    at :math:`\\infty`, we flip the stereographic projection axis and evaluate the latter polynomial at 0
    (and then complex conjugate). Insofar as pole flipping causes the highest degree term to become the lowest degree/constant term,
    evaluating at :math:`\\infty` amounts to returning the first coefficient.

    If `homogeneous=True`, returns a function which takes a spinor (as an nd.array, qt.Qobj, or two separate complex coordinates)
    and evaluates the homogeneous Majorana polynomial:

     .. math::

        p(z, w) = \\sum_{m=-j}^{m=j} (-1)^{j+m} \\sqrt{\\frac{(2j)!}{(j-m)!(j+m)!}} a_{j+m} w^{j-m} z^{j+m}

    (N.b. currently the normalization is off for the homogeneous case, but the correct roots are obtained.)

    If `cartesian=True`, returns a function with takes cartesian coordinates (as an nd.array or three sepaparate real components)
    and evaluates the Majorana polynomial by first converting the cartesian coordinates to an extended complex coordinate.

    If `spherical=True`, returns a function with takes spherical coordinates (as an nd.array or three sepaparate real components)
    and evaluates the Majorana polynomial by first converting the spherical coordinates to an extended complex coordinate.

    If `normalized=True`, returns the normalized versions of any of the above functions. Note that the normalized
    versions are no longer analytic/holomorphic. Given a extended complex coordinate :math:`z = re^{i\\theta}` (or :math:`\\infty`),
    the normalization factor is:

    .. math::

        \\frac{e^{-2ij\\theta}}{(1+r^2)^j}

    If :math:`z=\\infty`, again since we flip the poles, we use :math:`z=0`.
    When normalized, the resulting functions are equivalent to:

    .. math::
        \\langle -xyz \\mid \\psi \\rangle

    Where :math:`\\mid xyz \\rangle` refers to the spin coherent state which has all its "stars" at 
    cartesian coordinates :math:`x, y, z`, and :math:`\\mid \\psi \\rangle` is the spin in the :math:`\\mid j, m\\rangle`
    representation. In other words, evaluating a normalized Majorana function at :math:`x, y, z` is equivalent to evaluating:

    .. code-block::

       spin_coherent(j, -xyz).dag()*spin

    Which is the inner product between the spin and the spin coherent state antipodal to :math:`x, y, z`
    on the sphere. Since the Majorana stars are zeros of this function, we can interpret them
    as picking out those directions for which there's 0 probability that all the angular momentum is concentrated
    in the opposite direction. Insofar as we can think of each star as contributing a quantum of angular momentum :math:`\\frac{1}{2}`
    in that direction, naturally there's no chance that *all* the angular momentum is concentrated opposite to any of those points.
    By the fundamental theorem of algebra, knowing these points is equivalent to knowing the entire quantum state.

    Parameters
    ----------
    spin : qt.Qobj
        Spin-j state.

    projective : bool, optional
        Whether to return Majorana polynomial as a function of an extended complex coordinate.

    homogeneous : bool, optional
        Whether to return Majorana polynomial as a function of a spinor.

    cartesian : bool, optional
        Whether to return Majorana polynomial as a function of cartesian coordinates on unit sphere.

    spherical : bool, optional
        Whether to return Majorana polynomial as a function of spherical coordinates.

    normalize : bool, optional
        Whether to normalize the above functions.

    Returns
    -------
    poly : np.ndarray or func
        Either 2j+1 Majorana polynomial coefficients or else one of the above functions.
    """
    j = (spin.shape[0]-1)/2
    v = components(spin)
    poly = np.array(\
            [v[int(m+j)]*\
                (((-1)**(int(m+j)))*\
                np.sqrt(factorial(2*j)/(factorial(j-m)*factorial(j+m))))
                    for m in np.arange(-j, j+1)])
    if projective or cartesian or spherical:
        def __poly__(z):
            normalization = np.exp(-2j*j*np.angle(z))/(1+abs(z if z != np.inf else 0)**2)**j if normalized else 1
            return normalization*(poly[0] if z == np.inf else \
                        sum([poly[int(j+m)]*z**(j-m) for m in np.arange(-j, j+1)]))
        if projective:
            return __poly__
        if cartesian:
            def __cartesian__(*args):
                xyz = args[0] if len(args) == 1 else np.array(args)
                return __poly__(xyz_c(xyz))
            return __cartesian__
        if spherical:
            def __spherical__(*args):
                sph = args[0] if len(args) == 1 else np.array(args)
                return __poly__(sph_c(sph))
            return __spherical__
    if homogeneous: 
        def __hompoly__(*args):
            z, w = components(args[0]) if len(args) == 1 else args
            return sum([poly[int(j+m)]*(w**(j-m))*(z**(j+m))\
                        for m in np.arange(-j, j+1)])
        return __hompoly__
    return poly

def poly_spin(poly):
    """
    Converts a Majorana polynomial into a spin-j state.

    Parameters
    ----------
        poly : np.ndarray
            2j+1 Majorana polynomial coefficients.

    Returns
    -------
        spin : qt.Qobj
            Spin-j state.
    """
    j = (len(poly)-1)/2
    return qt.Qobj(np.array(\
            [poly[int(m+j)]/\
                (((-1)**(int(m+j)))*\
                np.sqrt(factorial(2*j)/(factorial(j-m)*factorial(j+m))))
                    for m in np.arange(-j, j+1)])).unit()

def poly_roots(poly):
    """
    Takes a Majorana polynomial to its roots. We use numpy's polynomial solver. 
    The number of initial coefficients which are 0 are intepreted as the number of
    roots at :math:`\\infty`. In other words, to the extent that the degree of 
    a Majorana polynomial corresponding to a spin-j state is less than 2j+1, we add
    that many roots at :math:`\\infty`.

    Parameters
    ----------
        poly : np.ndarray
            2j+1 Majorana polynomial coefficients.

    Returns
    -------
        roots : list
            Roots on the extended complex plane.

    """
    return [np.inf]*np.flatnonzero(poly)[0] + [complex(root) for root in np.roots(poly)]

def roots_poly(roots):
    """
    Takes a set of points on the extended complex plane and forms the polynomial 
    which has these points as roots. Roots at :math:`\\infty` turn into initial
    zero coefficients.

    Parameters
    ----------
        roots : list
            Roots on the extended complex plane.

    Returns
    -------
        poly : np.ndarray
            2j+1 Majorana polynomial coefficients.

    """
    zeros = roots.count(0j)
    if zeros == len(roots):
        return np.array([1] + [0j]*len(roots))
    poles = roots.count(np.inf)
    if poles == len(roots):
        return np.array([0j]*poles + [1])
    roots = [root for root in roots if root != np.inf]
    coeffs = np.array([((-1)**(-i))*sum([np.prod(terms)\
                        for terms in combinations(roots, i)])\
                            for i in range(len(roots)+1)])
    return np.concatenate([np.zeros(poles), coeffs])

def spin_xyz(spin):
    """
    Takes a spin-j state and returns the cartesian coordinates on the unit sphere
    corresponding to its "Majorana stars." Each contributes a quantum of angular 
    momentum :math:`\\frac{1}{2}` to the overall spin.

    Parameters
    ----------
        spin : qt.Qobj
            Spin-j state.

    Returns
    -------
        xyz : np.ndarray
            A array with shape (2j, 3) containing the 2j cartesian coordinates of the stars.
    """
    return np.array([c_xyz(root) for root in poly_roots(spin_poly(spin))])

def xyz_spin(xyz):
    """
    Given the cartesian coordinates of a set of "Majorana stars," returns the
    corresponding spin-j state, which is defined only up to complex phase.

    Parameters
    ----------
        xyz : np.ndarray
            A array with shape (2j, 3) containing the 2j cartesian coordinates of the stars.

    Returns
    -------
        spin : qt.Qobj
            Spin-j state.
    """
    return poly_spin(roots_poly([xyz_c(star) for star in xyz]))

def spin_spinors(spin):
    """
    Takes a spin-j state and returns its decomposition into 2j spinors.

    Parameters
    ----------
        spin : qt.Qobj
            Spin-j state.

    Returns
    -------
        spinors : list
            2j spinors.
    """
    return [c_spinor(c) for c in poly_roots(spin_poly(spin))]

def spinors_spin(spinors):
    """
    Given 2j spinors returns the corresponding spin-j state (up to phase).

    Parameters
    ----------
        spinors : list
            2j spinors.

    Returns
    -------
        spin : qt.Qobj
            Spin-j state.
    """
    return poly_spin(roots_poly([spinor_c(spinor) for spinor in spinors]))

def spin_c(spin):
    """
    Takes a spin-j state and returns its decomposition into 2j roots on the extended complex plane.

    Parameters
    ----------
        spin : qt.Qobj
            Spin-j state.

    Returns
    -------
        c : list
            2j extended complex roots.
    """
    return poly_roots(spin_poly(spin))

def c_spin(c):
    """
    Takes 2j roots on the extended complex plane and returns the corresponding spin-j state (up to complex phase).

    Parameters
    ----------
        c : list
            2j extended complex roots.

    Returns
    -------
        spin : qt.Qobj
            Spin-j state.
    """
    return poly_spin(roots_poly(c))

def spin_sph(spin):
    """
    Takes a spin-j state and returns its decomposition into 2j "stars" given in spherical coordinates.

    Parameters
    ----------
        spin : qt.Qobj
            Spin-j state.

    Returns
    -------
        sph : np.array
            A array with shape (2j, 3) containing the 2j spherical coordinates of the stars.
    """
    return np.array([c_sph(c) for c in poly_roots(spin_poly(spin))])

def sph_spin(sph):
    """
    Takes 2j "stars" given in spherical coordinates and returns the corresponding spin-j state (up to complex phase).

    Parameters
    ----------
        sph : np.array
            A array with shape (2j, 3) containing the 2j spherical coordinates of the stars.

    Returns
    -------
        spin : qt.Qobj
            Spin-j state.
    """
    return poly_spin(roots_poly([sph_c(s) for s in sph]))

def spin_coherent(j, coord, from_cartesian=True,\
                            from_spherical=False,\
                            from_complex=False,\
                            from_spinor=False):
    """
    Returns the spin-j coherent state which is defined by having
    all its Majorana stars located at a single point on the sphere. 
    This point can be given in terms of cartesian, spherical, extended complex, and spinorial coordinates.

    Parameters
    ----------
        j : int
            j value which indexes the :math:`SU(2)` representation.
        coord : nd.array or qt.Qobj or complex/inf
            Coordinates specifying the direction of the spin coherent state.
        from_cartesian : bool, optional
            Whether the provided coordinates are cartesian (default).
        from_spherical : bool, optional
            Whether the provided coordinates are spherical.
        from_complex : bool, optional
            Whether the provided coordinates are extended complex.
        from_spinor : bool, optional
            Whether the provided coordinates are spinorial.

    Returns
    -------
        spin_coherent : qt.Qobj
            Spin-j coherent state in the specified direction.
    """
    if from_cartesian:
        r, phi, theta = xyz_sph(coord)
    if from_spherical:
        r, phi, theta = coord
    if from_complex:
        r, phi, theta = c_sph(coord)
    if from_spinor:
        r, phi, theta = c_sph(spinor_c(coord))
    return qt.spin_coherent(j, theta, phi)

def antipodal(to_invert, from_cartesian=False,\
                         from_spherical=False):
    """
    If given an extended complex coordinate, takes the point to its antipode on the sphere via the map:
    
    .. math::

        z \\rightarrow -\\frac{z}{|z|^2} = - \\frac{1}{z^{*}}

    If :math:`z=\\infty`, :math:`z \\rightarrow 0` and if :math:`z=0`, :math:`z \\rightarrow \\infty`.

    If given a spin state or polynomial coefficients, inverts the whole sphere.
    This could be done by inverting the individual roots, or directly on the state/polynomial by
    reversing the components, complex conjugating, and multiplying every other
    component by :math:`-1`.

    If `from_cartesian=True` or `from_spherical=True`, the argument is interpreted
    as a single coordinate in terms of those coordinate systems and the flipped coordinate 
    is returned in the same.

    Parameters
    ----------
        to_invert : complex/inf or qt.Qobj or np.ndarray
            Extended complex coordinate, cartesian coordinate,
            spherical coordinate, or spin state/polynomial to invert.
    Returns
    -------
        inverted : complex/inf or qt.Qobj or np.ndarray
            Inverted extended complex coordinate, cartesian coordinate,
            spherical coordinate or spin state/polynomial.
    """
    if np.isscalar(to_invert):
        if np.isclose(to_invert, 0):
            return np.inf
        if to_invert == np.inf:
            return 0
        return -to_invert/np.abs(to_invert)**2
    if from_cartesian:
        return -1*to_invert
    if from_spherical:
        return c_sph(antipodal(sph_c(to_invert)))
    inverted = np.array([c*(-1)**(i) for i, c in enumerate(components(to_invert)[::-1].conj())])
    return qt.Qobj(inverted) if type(to_invert) == qt.Qobj else inverted

def poleflip(to_flip, from_cartesian=False,\
                      from_spherical=False):
    """
    Flips the pole of projection.  If given an extended complex coordinate, this amounts to
    projecting to the sphere via a South Pole projection and then projecting back to the plane 
    via a North Pole projection.

    More simply:

    .. math::

        z \\rightarrow \\frac{z}{|z|^2} = \\frac{1}{z^{*}}

    If :math:`z=\\infty`, :math:`z \\rightarrow 0` and if :math:`z=0`, :math:`z \\rightarrow \\infty`.

    If given a spin state or polynomial coefficients, flips the pole of projection for the entire state.
    This could be done by flipping the individual roots, or directly on the state/polynomial by
    reversing the components and complex conjugating.

    This is useful for evaluating the Majorana polynomial at :math:`\\infty`. We actually
    need a second coordinate chart. We flip the projection pole and evaluate at :math:`0` instead.

    If `from_cartesian=True` or `from_spherical=True`, the argument is interpreted
    as a single coordinate in terms of those coordinate systems and the flipped coordinate 
    is returned in the same.

    Parameters
    ----------
        to_flip : (complex/inf) or qt.Qobj or np.ndarray
            Extended complex coordinate, cartesian coordinate,
            spherical coordinate, or spin state/polynomial to flip.
        from_cartesian : bool, optional
            Whether to interpret argument as cartesian coordinates.
        from_spherical : bool, optional
            Whether to interpret argument as spherical coordinates.

    Returns
    -------
        flipped : (complex/inf) or qt.Qobj or np.ndarray
            Flipped extended complex coordinate, cartesian coordinate,
            spherical coordinate or spin state/polynomial.
    """
    if np.isscalar(to_flip):
        if np.isclose(to_flip, 0):
            return np.inf
        if to_flip == np.inf:
            return 0
        return to_flip/np.abs(to_flip)**2
    if from_cartesian:
        x, y, z = to_flip
        return np.array([x, y, -z])
    if from_spherical:
        return c_sph(poleflip(sph_c(to_flip)))
    flipped = components(to_flip)[::-1].conj()
    return qt.Qobj(flipped) if type(to_flip) == qt.Qobj else flipped

def mobius(abcd):
    """
    Given parameters :math:`\\begin{pmatrix} a & b \\\\ b & c \\end{pmatrix}`,
    arranged in a 2x2 matrix, returns a function which implements the 
    corresponding `Möbius transformation <https://en.wikipedia.org/wiki/M%C3%B6bius_transformation>`_

    .. math::

        f(z) = \\frac{az+b}{cz+d}

    which acts on the extended complex plane. Note that if :math:`c \\neq 0`, we have:

    .. math::

        f(-\\frac{d}{c}) = \\infty

        f(\\infty) = \\frac{a}{c}

    And if :math:`c = 0`, we have:

    .. math::

        f(\\infty) = \\infty

    Parameters
    ----------
        abcd : np.ndarray or qt.Qobj
            2x2 matrix representing Möbius parameters.

    Returns
    -------
        mobius : func
            A function which takes an extended complex coordinate as input, 
            and returns an extended complex coordinate as output.

    Raises
    ------
        Exception
            If :math:`ad = bc`, then :math:`f(z) = \\frac{a}{c}`, a constant function,
            which doesn't qualify as a Möbius transformation.
    """
    if type(abcd) == qt.Qobj:
        abcd = abcd.full()
    a, b, c, d = abcd.reshape(4)
    if a*d - b*c == 0:
        raise Exception("Möbius transformation doesn't satisfy ad-bc != 0.")
    def __mobius__(z):
        if c != 0:
            if z == -d/c:
                return np.inf
            if z == np.inf:
                return a/c
        else:
            if z == np.inf:
                return np.inf
        return (a*z + b)/(c*z + d)
    return __mobius__

def xyz_eigenstates(j, m, direction):
    """
    Returns eigenstates of Pauli operators.

    Parameters
    ----------
        j : float
            j value of representation.
        m : float
            m value of representation.
        direction : str
            "x", "y", or "z".
    """
    if direction == "x":
        up = np.array([1,0,0])
        down = np.array([-1,0,0])
    elif direction == "y":
        up = np.array([0,1,0])
        down = np.array([0,-1,0])
    elif direction == "z":
        up = np.array([0,0,1])
        down = np.array([0,0,-1])
    nup, ndown = [(int(2*j-i), i)\
                    for i in range(int(2*j+1))]\
                        [list(np.arange(j, -j-1, -1)).index(m)]
    return xyz_spin([up]*nup + [down]*ndown)

